/* Parser for Document Structure Files */

%{                       /* C declarations */
#include "opldefs.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#ifndef __GNUC__
#define alloca malloc
#endif

#if defined(WIN32) || defined(__GNUC__)
#define xmalloc malloc
#endif

#define YYDEBUG 1
#define YYERROR_VERBOSE 1
extern int yylex( void );
void yyerror( char* s );
extern int _lineNo;

FILE *in;
%}

/*%pure_parser*/

%union
{
  char           *string;
  int             integer;
  ident          *id;
  identList      *idlist;
  procedure      *proc;
  procedureList  *plist;
  statement      *stmt;
  statementList  *slist;
  expression     *expr;
  expressionList *elist;
  program        *prog;
}

/* Parsable tokens */

%token <integer>  INTEGER HEXINT CHARACTER
%token <id>       IDENTIFIER SUBROUTINE
%token <string>   STRING
%token <id>       OPL_FUNCTION
%token <string>   OPL_PROCEDURE OPL_PROCEDURE_NO_ARGS OPL_PRINT_PROC

%token APP AND DO ELSE ELSEIF ENDA ENDP ENDIF ENDWH
%token GLOBAL IF LOCAL OR PROC RETURN UNTIL WHILE

%token LE, GE, NE, POWER     /* <=, >=, <>, ** */


/* Return Types of the non-terminals */

%type <prog>    program
%type <plist>   opt_procedure_list
%type <proc>    procedure
%type <idlist>  opt_argument_declaration_list, opt_identifier_declaration_list
%type <id>      identifier_declaration
%type <slist>   opt_statement_list, elseif_statement_list
%type <stmt>    statement
%type <elist>   opt_expression_list, expression_list, expression_list_tail
%type <elist>   opt_print_expression_list, print_expression_list_tail
%type <elist>   opt_array_declaration
%type <expr>    expression, primary_expression, multiplicative_expression
%type <expr>    additive_expression, relational_expression, equality_expression
%type <expr>    logical_or_expression, unary_expression

/* Grammar follows */

%%

/*******************************************************************************
 *                               Program                                       *
 ******************************************************************************/

program:
    opt_app_block procedure opt_procedure_list
      { prog.pProcs = new_procedureList( $2, $3 ); }
  ;

opt_app_block:
    /* empty */
  | APP ENDA
  ;

/*******************************************************************************
 *                             Procedures                                      *
 ******************************************************************************/

opt_procedure_list:
    /* empty */
      { $$ = NULL; }
  | procedure opt_procedure_list
      { $$ = new_procedureList( $1, $2 ); }
  ;

procedure:
    PROC SUBROUTINE opt_argument_declaration_list opt_statement_list ENDP
      { $$ = new_procedure( $2, $3, $4 ); }
  ;

opt_argument_declaration_list:
    /* empty */
      { $$ = NULL; }
  | '(' identifier_declaration opt_identifier_declaration_list ')'
      { $$ = new_identList( $2, $3 ); }
    ;

/*******************************************************************************
 *                             Statements                                      *
 ******************************************************************************/

opt_identifier_declaration_list:
    /* empty */
      { $$ = NULL; }
  | ',' identifier_declaration opt_identifier_declaration_list
      { $$ = new_identList( $2, $3 ); }
    ;

identifier_declaration:
    IDENTIFIER opt_array_declaration
      {
        $1->pArrays = $2;
        $$ = $1;
      }
  ;

opt_array_declaration:
    /* empty */
      { $$ = NULL; }
  | '(' expression_list ')'
      { $$ = $2; }
  ;

opt_statement_list:
    /* empty */
      { $$ = NULL; }
  | ':' opt_statement_list
      { $$ = $2; }
  | statement opt_statement_list
      { $$ = new_statementList( $1, $2 ); }
    ;

elseif_statement_list:
    ENDIF
      { $$ = NULL; }
  | ELSE opt_statement_list ENDIF
      { $$ = $2; }
  | ELSEIF expression opt_statement_list elseif_statement_list
      { $$ = new_statementList( new_if_statement( $2, $3, $4 ), NULL ); }
    ;

statement:
    IF expression opt_statement_list elseif_statement_list
      { $$ = new_if_statement( $2, $3, $4 ); }
  | WHILE expression opt_statement_list ENDWH
      { $$ = new_while_statement( $2, $3 ); }
  | DO opt_statement_list UNTIL expression
      { $$ = new_do_statement( $4, $2 ); }
  | IDENTIFIER '(' opt_expression_list ')' '=' expression
      {
        $1->pArrays = $3;
        $$ = new_assignment_statement( $1, $6 );
      }
  | IDENTIFIER '=' expression
      { $$ = new_assignment_statement( $1, $3 ); }
  | GLOBAL identifier_declaration opt_identifier_declaration_list
      { $$ = new_global_statement( new_identList( $2, $3 ) ); }
  | LOCAL identifier_declaration opt_identifier_declaration_list
      { $$ = new_local_statement( new_identList( $2, $3 ) ); }
  | OPL_PRINT_PROC opt_print_expression_list
      { $$ = new_call_statement( new_ident( $1, untyped_id, NULL ), $2, 1 ); }
  | OPL_PROCEDURE_NO_ARGS
      { $$ = new_call_statement( new_ident( $1, untyped_id, NULL ), NULL, 1 ); }
  | OPL_PROCEDURE expression_list
      { $$ = new_call_statement( new_ident( $1, untyped_id, NULL ), $2, 1 ); }
  | RETURN
      { $$ = new_return_statement( NULL ); }
  | RETURN expression
      { $$ = new_return_statement( $2 ); }
  | SUBROUTINE '(' expression_list ')'
      { $$ = new_call_statement( $1, $3, 0 ); }
  | SUBROUTINE
      { $$ = new_call_statement( $1, NULL, 0 ); }
  ;

/*******************************************************************************
 *                             Expressions                                     *
 ******************************************************************************/

opt_expression_list:
    /* empty */
      { $$ = NULL; }
  | expression_list
      { $$ = $1; }
  ;

expression_list:
    expression expression_list_tail
      { $$ = new_expressionList( $1, $2 ); }
    ;

expression_list_tail:
    /* empty */
      { $$ = NULL; }
  | ',' expression expression_list_tail
      { $$ = new_expressionList( $2, $3 ); }
  ;

opt_print_expression_list:
    /* empty */
      { $$ = NULL; }
  | expression print_expression_list_tail
      { $$ = new_expressionList( $1, $2 ); }
  ;

print_expression_list_tail:
    /* empty */
      { $$ = NULL; }
  | ',' expression print_expression_list_tail
      { $$ = new_expressionList( $2, $3 ); }
  | ';' expression print_expression_list_tail
      { $$ = new_expressionList( $2, $3 ); }
  /* | ','
      { $$ = NULL; } */
  | ';'
      { $$ = NULL; }
  ;

expression:
    logical_or_expression
  | expression AND logical_or_expression
      { $$ = new_binary_op_expression( and_expr, $1, $3 ); }
  ;

logical_or_expression:
    equality_expression
  | logical_or_expression OR equality_expression
      { $$ = new_binary_op_expression( or_expr, $1, $3 ); }
  ;

equality_expression:
    relational_expression
  | equality_expression '=' relational_expression
      { $$ = new_binary_op_expression( equality_expr, $1, $3 ); }
  | equality_expression NE relational_expression
      { $$ = new_binary_op_expression( inequality_expr, $1, $3 ); }
  ;

relational_expression:
    additive_expression
  | relational_expression '<' additive_expression
      { $$ = new_binary_op_expression( lessthan_expr, $1, $3 ); }
  | relational_expression '>' additive_expression
      { $$ = new_binary_op_expression( greaterthan_expr, $1, $3 ); }
  | relational_expression LE additive_expression
      { $$ = new_binary_op_expression( lessequal_expr, $1, $3 ); }
  | relational_expression GE additive_expression
      { $$ = new_binary_op_expression( greaterequal_expr, $1, $3 ); }
  ;

additive_expression:
    multiplicative_expression
  | additive_expression '+' multiplicative_expression
      { $$ = new_binary_op_expression( addition_expr, $1, $3 ); }
  | additive_expression '-' multiplicative_expression
      { $$ = new_binary_op_expression( subtraction_expr, $1, $3 ); }
  ;

multiplicative_expression:
    unary_expression
  | multiplicative_expression '*' unary_expression
      { $$ = new_binary_op_expression( multiplication_expr, $1, $3 ); }
  | multiplicative_expression '/' unary_expression
      { $$ = new_binary_op_expression( division_expr, $1, $3 ); }
  | multiplicative_expression POWER unary_expression
      { $$ = new_binary_op_expression( power_expr, $1, $3 ); }
  ;

unary_expression:
    primary_expression
      { $$ = $1; }
  | '-' primary_expression
      { $$ = new_unary_op_expression( uminus_expr, $2 ); }
  ;

primary_expression:
    INTEGER
      { $$ = new_integer_expr( integer_expr, $1 ); }
  | HEXINT
      { $$ = new_integer_expr( hexint_expr, $1 ); }
  | CHARACTER
      { $$ = new_integer_expr( character_expr, $1 ); }
  | IDENTIFIER '(' opt_expression_list ')'
      {
        $1->pArrays = $3;
        $$ = new_variable_expr( $1 );
      }
  | IDENTIFIER
      { $$ = new_variable_expr( $1 ); }
  | STRING
      { $$ = new_string_expr( $1 ); }
  | OPL_FUNCTION '(' expression_list ')'
      { $$ = new_call_expression( $1, $3, 1 ); }
  | OPL_FUNCTION
      { $$ = new_call_expression( $1, NULL, 1 ); }
  | SUBROUTINE '(' expression_list ')'
      { $$ = new_call_expression( $1, $3, 0 ); }
  | SUBROUTINE
      { $$ = new_call_expression( $1, NULL, 0 ); }
  | '(' expression ')'
      { $$ = new_unary_op_expression( parenthesis_expr, $2 ); }
  ;

%%
